package dev.shadowsoffire.apotheosis.adventure.loot;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import com.google.common.base.Preconditions;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableList;
import dev.shadowsoffire.apotheosis.Apotheosis;
import dev.shadowsoffire.apotheosis.adventure.AdventureModule;
import dev.shadowsoffire.apotheosis.adventure.affix.Affix;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixRegistry;
import dev.shadowsoffire.apotheosis.adventure.affix.AffixType;
import dev.shadowsoffire.apotheosis.adventure.loot.LootCategory;
import dev.shadowsoffire.apotheosis.adventure.loot.LootRarity;
import dev.shadowsoffire.placebo.codec.CodecProvider;
import dev.shadowsoffire.placebo.reload.DynamicHolder;
import dev.shadowsoffire.placebo.reload.WeightedDynamicRegistry;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.Items;

public class RarityRegistry extends WeightedDynamicRegistry<LootRarity> {
  public static final dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry INSTANCE = new dev.shadowsoffire.apotheosis.adventure.loot.RarityRegistry();
  
  protected List<DynamicHolder<LootRarity>> ordered = new ArrayList<>();
  
  protected BiMap<Item, DynamicHolder<LootRarity>> materialMap = HashBiMap.create();
  
  private RarityRegistry() {
    super(AdventureModule.LOGGER, "rarities", true, false);
  }
  
  public static boolean isMaterial(Item item) {
    return getMaterialRarity(item).isBound();
  }
  
  public static DynamicHolder<LootRarity> getMaterialRarity(Item item) {
    return INSTANCE.materialMap.getOrDefault(item, INSTANCE.emptyHolder());
  }
  
  public static DynamicHolder<LootRarity> getMinRarity() {
    return INSTANCE.ordered.get(0);
  }
  
  public static DynamicHolder<LootRarity> getMaxRarity() {
    return INSTANCE.ordered.get(INSTANCE.ordered.size() - 1);
  }
  
  public static DynamicHolder<LootRarity> byOrdinal(int i) {
    return INSTANCE.ordered.get(i);
  }
  
  public static ResourceLocation convertId(String s) {
    return s.contains(":") ? new ResourceLocation(s) : Apotheosis.loc(s);
  }
  
  public static DynamicHolder<LootRarity> byLegacyId(String s) {
    return INSTANCE.holder(convertId(s));
  }
  
  public static DynamicHolder<LootRarity> prev(DynamicHolder<LootRarity> rarity) {
    if (rarity == getMinRarity())
      return rarity; 
    return byOrdinal(((LootRarity)rarity.get()).ordinal() - 1);
  }
  
  public static DynamicHolder<LootRarity> next(DynamicHolder<LootRarity> rarity) {
    if (rarity == getMaxRarity())
      return rarity; 
    return byOrdinal(((LootRarity)rarity.get()).ordinal() + 1);
  }
  
  public List<DynamicHolder<LootRarity>> getOrderedRarities() {
    return this.ordered;
  }
  
  protected void beginReload() {
    super.beginReload();
    this.ordered = new ArrayList<>();
    this.materialMap = HashBiMap.create();
  }
  
  protected void onReload() {
    super.onReload();
    this.ordered = this.registry.values().stream().sorted(Comparator.comparing(LootRarity::ordinal)).map(this::holder).toList();
    int lastOrdinal = -1;
    for (DynamicHolder<LootRarity> r : this.ordered) {
      if (((LootRarity)r.get()).ordinal() != lastOrdinal + 1) {
        StringBuilder sb = new StringBuilder();
        sb.append("Rarity ordinal order is inconsistent. The ordinals must start at zero and be continuous up to the max value.\n");
        for (DynamicHolder<LootRarity> rarity : this.ordered)
          sb.append("" + rarity.getId() + " | " + rarity.getId() + "\n"); 
        throw new RuntimeException(sb.toString());
      } 
      lastOrdinal = ((LootRarity)r.get()).ordinal();
    } 
    for (DynamicHolder<LootRarity> r : this.ordered) {
      DynamicHolder<LootRarity> old = this.materialMap.put(((LootRarity)r.get()).getMaterial(), r);
      if (old != null)
        throw new RuntimeException("Two rarities may not share the same rarity material: " + r.getId() + " conflicts with " + old.getId()); 
    } 
    this.ordered = ImmutableList.copyOf(this.ordered);
  }
  
  protected void registerBuiltinCodecs() {
    registerDefaultCodec(Apotheosis.loc("rarity"), LootRarity.LOAD_CODEC);
  }
  
  protected void validateItem(ResourceLocation key, LootRarity item) {
    super.validateItem(key, (CodecProvider)item);
    Preconditions.checkNotNull(item.getColor());
    Preconditions.checkArgument((item.getMaterial() != null && item.getMaterial() != Items.f_41852_));
    Preconditions.checkArgument((item.getWeight() >= 0), "A rarity may not have negative weight!");
    Preconditions.checkArgument((item.getQuality() >= 0.0F), "A rarity may not have negative quality!");
    Preconditions.checkArgument(!item.getRules().isEmpty(), "A rarity may not have no rules!");
  }
  
  public void validateLootRules() {
    for (Iterator<LootRarity> iterator = this.registry.values().iterator(); iterator.hasNext(); ) {
      LootRarity rarity = iterator.next();
      Map<AffixType, List<LootRarity.LootRule>> sorted = new HashMap<>();
      rarity.getRules().stream().filter(r -> r.type().needsValidation()).forEach(rule -> {
            sorted.computeIfAbsent(rule.type(), ());
            ((List<LootRarity.LootRule>)sorted.get(rule.type())).add(rule);
          });
      sorted.forEach((type, rules) -> {
            for (LootCategory cat : LootCategory.VALUES) {
              if (cat.isNone())
                continue; 
              List<Affix> affixes = AffixRegistry.INSTANCE.getValues().stream().filter(()).toList();
              if (affixes.size() < rules.size()) {
                StringBuilder errMsg = new StringBuilder();
                errMsg.append("Insufficient number of affixes to satisfy the loot rules (ignoring backup rules) of rarity " + getKey((CodecProvider)rarity) + " for category " + cat.getName());
                errMsg.append("Required: " + rules.size());
                errMsg.append("; Provided: " + affixes.size());
                AdventureModule.LOGGER.warn(errMsg.toString());
              } 
            } 
          });
    } 
  }
}
